/* bioshd.c */
/* Copyright (C) 1994 Yggdrasil Computing, Incorporated
   4880 Stevens Creek Blvd. Suite 205
   San Jose, CA 95129-1034
   USA
   Tel (408) 261-6630
   Fax (408) 261-6631

   This file is part of the Linux Kernel

   Linux is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Linux is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Linux; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include <linuxmt/types.h>
#include <linuxmt/kernel.h>
#include <linuxmt/sched.h>
#include <linuxmt/errno.h>
/*#include <linux/head.h>*/
#include <linuxmt/genhd.h>
#include <linuxmt/hdreg.h>
#include <linuxmt/biosparm.h>
#include <linuxmt/major.h>
#include <linuxmt/bioshd.h>
#include <linuxmt/fs.h>
/*#include <linux/signal.h>*/
#include <linuxmt/string.h>
#include <arch/segment.h>

#define MAJOR_NR BIOSHD_MAJOR
#define BIOSDISK
#include "blk.h"

static int bioshd_ioctl ();
static int  bioshd_open ();
static void  bioshd_release ();

static struct file_operations bioshd_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* select */
	bioshd_ioctl,			/* ioctl */
	NULL,			/* mmap */
	bioshd_open,		/* open */
	bioshd_release,		/* release */
};

static struct wait_queue * busy_wait = NULL;

static int bioshd_initialized=0;
static int force_bioshd;
static int revalidate_hddisk();
static struct {
  int cylinders;
  int sectors;
  int heads;
} drive_info[4];

static struct hd_struct hd[4<<6];
static char busy[4] = {0, };
static int access_count[4] = {0, };
static unsigned char hd_drive_map[4]={
	0x80,0x81,	/* hda, hdb */
	0x00,0x01	/* fda, fdb */
};

static void bioshd_geninit();

static int bioshd_sizes[4<<6] = {0, };

static struct gendisk bioshd_gendisk = {
	MAJOR_NR,	/* Major number */	
	"hd",		/* Major name */
	6,		/* Bits to shift to get real from partition */
	1 << 6,		/* Number of partitions per real */
	4,		/* maximum number of real */
	bioshd_geninit,	/* init function */
	hd,		/* hd struct */
	bioshd_sizes,/* sizes not blocksizes*/
	0,		/* number */
	(void *) drive_info,	/* internal */
	NULL		/* next */
};

static void
bioshd_release (inode,filp)
struct inode *inode;
struct file *filp;
{
  int target;
  sync_dev(inode->i_rdev);

  target =  DEVICE_NR(MINOR(inode->i_rdev));
  access_count[target]--;
  return;
}

static int
bioshd_open (inode,filp)
struct inode *inode;
struct file *filp;
{
  int minor;
  int target;

  minor=MINOR(inode->i_rdev)>>6;
  if (bioshd_initialized == 0) {
    return (-ENXIO);
  }
  if (minor >= 4) {
    return (-ENXIO);
  }
  if (hd[MINOR(inode->i_rdev)].start_sect == -1) {
    return (-ENXIO);
  }

  target =  DEVICE_NR(MINOR(inode->i_rdev));

  while (busy[target])
    sleep_on(&busy_wait);
  access_count[target]++;
  return 0;

}
  
void init_bioshd()
{     
	int i;
	int count;
	int hdcount;
	struct gendisk *ptr;

	printk("hd Driver Copyright (C) 1994 Yggdrasil Computing, Inc.\n");
	printk("Extended and modified for Linux8086 by Alan Cox.\n");

	/*
	 *	Now we need to check to see if the driver is in place and
	 *	determine how many units we are dealing with.
	 */

	for (i=0; i < 2; i++) 
	{
		drive_info[i].heads=-1;
		BD_IRQ = BIOSHD_INT;
		BD_AX = BIOSHD_TESTDRIVE;
		BD_DX = 0x80+i;
		call_bios();
		if (CARRY_SET || (BD_AX>>8) != 0) 
			continue;

		BD_IRQ = BIOSHD_INT;
		BD_AX = BIOSHD_DRIVE_PARMS;
		BD_DX = hd_drive_map[i];
		call_bios();
		if (CARRY_SET || (BD_AX>>8) != 0) 
			continue;
		drive_info[i].heads = (BD_DX>>8)+1;
		drive_info[i].sectors = (BD_CX&0x3f);
		drive_info[i].cylinders = (BD_CX>>8)+((BD_CX&0xc0)<<2)+1;
	}

	/*
	 *	Hack hack, hardcode a 1.44Mb floppy drive
	 */

	drive_info[2].heads=2;
	drive_info[2].sectors=18;
	drive_info[2].cylinders=80;
	drive_info[3].heads=-1;
     
	count=0;
	hdcount=0;

	for (i=0; i < 4; i++) 
	{
		if (drive_info[i].heads != -1) 
		{
			count++;
			if(i<2)
				hdcount++;
		}
	}

	if (!count)
		return;

	if (count == 1) 
	     	printk("hd: found 1 drive\n");
	else
		printk("hd: found %d drives\n", count);

	for (i=0; i < 2; i++) 
	{
		if (drive_info[i].heads != -1) 
		{
			printk ("/dev/hd%c: %d heads, %d cylinders, %d sectors = %ld Meg\n",
				i+'a',
				drive_info[i].heads,
				drive_info[i].cylinders,
				drive_info[i].sectors,
				(long)drive_info[i].heads*drive_info[i].cylinders*
					drive_info[i].sectors*512L/(1024L*1024L));
		}
	}
	for(i=2;i<4;i++)
	{
		if(drive_info[i].heads != -1)
		{
			printk("/dev/fd%c: %d heads, %d cylinders, %d sectors = %ld K\n",
				i+'0'-2,
				drive_info[i].heads,
				drive_info[i].cylinders,
				drive_info[i].sectors,
				(long)drive_info[i].heads*drive_info[i].cylinders*
					drive_info[i].sectors*512L/1024L);
		}
	}

	bioshd_gendisk.nr_real = hdcount;
	i=register_blkdev(MAJOR_NR, DEVICE_NAME, &bioshd_fops);

	if (i==0) 
	{
		blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
		blksize_size[MAJOR_NR]=1024;
		read_ahead[MAJOR_NR]=2;
		if (gendisk_head == NULL) 
		{
			bioshd_gendisk.next=gendisk_head;
			gendisk_head = &bioshd_gendisk;
		}
		else
		{
			for (ptr = gendisk_head; ptr->next != NULL; ptr=ptr->next);
				ptr->next = &bioshd_gendisk;
			bioshd_gendisk.next = NULL;
		}
		bioshd_initialized=1;
		return;
	}

	printk ("hd: unable to register");
	return;
}


/* As far as I can tell this doesn't actually
   work, but we might as well try it. 
   -- Some XT controllers are happy with it.. [AC]
*/

static void reset_bioshd(minor)
int minor;
{
	BD_IRQ = BIOSHD_INT;
	BD_AX = BIOSHD_RESET;
	BD_DX = hd_drive_map[(minor>>6)];

	call_bios();
	if (CARRY_SET || (BD_AX&0xff00) != 0)
		printk ("hd: unable to reset.\n");
	return;
}


static int bioshd_ioctl(inode,file,cmd,arg)
struct inode * inode;
struct file * file;
unsigned int cmd;
unsigned long arg;
{
	return -EINVAL;
}


static void do_bioshd_request() 
{
	unsigned long count;
	unsigned long this_pass;
	short cylinder;
	short head;
	short sector;
	unsigned long start;
	int tmp;
	char *buff;
	int minor;
	int errs;
	int part;
	int drive;

	while (1) 
	{
next_block:

		/* make sure we have a valid request */
		if (!CURRENT || CURRENT->dev < 0)
			return;

		/* now initialize it */
		INIT_REQUEST;

		/* make sure it's still valid */
		if (CURRENT == NULL || CURRENT->sector == -1)
			return;
	
		if (bioshd_initialized!=1) 
		{
			end_request(0);
			continue;
		}
	  
		minor=MINOR(CURRENT->dev);
		part=minor&((1<<6)-1);
		drive =minor >> 6;
	
		/* make sure it's a disk we are dealing with. */
		if (drive>3 || drive < 0 || drive_info[drive].heads==-1) 
		{
			printk ("hd: non-existant drive\n");
			end_request(0);
			continue;
		}
	
		count=CURRENT->nr_sectors;
		start=CURRENT->sector;
		buff=CURRENT->buffer;
	
		if (hd[minor].start_sect== -1 || start >= hd[minor].nr_sects) 
		{
			printk ("hd: bad partition start=%d sect=%d nr_sects=%d.\n",
				start, (int)hd[minor].start_sect, (int)hd[minor].nr_sects);
			end_request(0);
			continue;
		}
	
		start+=hd[minor].start_sect;
		errs=0;

		while (count > 0) 
		{
			sector=(start%drive_info[drive].sectors)+1;
			tmp = start/drive_info[drive].sectors;
			head=tmp%drive_info[drive].heads;
			cylinder=tmp/drive_info[drive].heads;
	
			this_pass=count;

			BD_IRQ=BIOSHD_INT;
			if (CURRENT->cmd==WRITE) 
				BD_AX=BIOSHD_WRITE|this_pass; 
			else
				BD_AX=BIOSHD_READ|this_pass; 
	
			BD_BX=buff;
			BD_ES=get_ds();
			BD_CX=(cylinder<<8)|((cylinder>>2) &0xc0)|sector;
			BD_DX=(head << 8)|hd_drive_map[drive];
			BD_FL=0;
#if 0      
			printk("cylinder=%d head=%d sector=%d drive=%d CMD=%d\n",
				cylinder,head,sector,drive, CURRENT->cmd);
			printk("blocks %d\n", this_pass);
#endif			
	
			sti();
			call_bios();

			if (CARRY_SET) 
			{
				printk ("hd: error: AX=%d\n",BD_AX);
				reset_bioshd(MINOR(CURRENT->dev));
				errs++;
				if (errs>MAX_ERRS) 
				{
					end_request(0);
					goto next_block;
				}
				continue; /* try again */
			}
			count -= this_pass;
			start -= this_pass;
			buff += this_pass*512;
		}
		/* satisfied that request */
		end_request(1);
	}
}

#define DEVICE_BUSY busy[target]
#define USAGE access_count[target]
#define CAPACITY (drive_info[target].heads*drive_info[target].sectors*drive_info[target].cylinders)
/* We assume that the the bios parameters do not change, so the disk capacity
   will not change */
#undef MAYBE_REINIT
#define GENDISK_STRUCT bioshd_gendisk

/*
 * This routine is called to flush all partitions and partition tables
 * for a changed cdrom drive, and then re-read the new partition table.
 * If we are revalidating a disk because of a media change, then we
 * enter with usage == 0.  If we are using an ioctl, we automatically have
 * usage == 1 (we need an open channel to use an ioctl :-), so this
 * is our limit.
 */
static int revalidate_hddisk(dev,maxusage)
int dev;
int maxusage;
{
	int target, major;
	struct gendisk * gdev;
	int max_p;
	int start;
	int i;

	target =  DEVICE_NR(MINOR(dev));
	gdev = &GENDISK_STRUCT;

	cli();
	if (DEVICE_BUSY  || USAGE > maxusage ) {
		sti();
		return -EBUSY;
	};
	DEVICE_BUSY = 1;
	sti();

	max_p = gdev->max_p;
	start = target << gdev->minor_shift;
	major = MAJOR_NR << 8;

	for (i=max_p - 1; i >=0 ; i--) {
		sync_dev(major | start | i);
		invalidate_inodes(major | start | i);
		invalidate_buffers(major | start | i);
		gdev->part[start+i].start_sect = 0;
		gdev->part[start+i].nr_sects = 0;
	};

#ifdef MAYBE_REINIT
	MAYBE_REINIT;
#endif

	gdev->part[start].nr_sects = CAPACITY;
	resetup_one_dev(gdev, target);

	DEVICE_BUSY = 0;
	wake_up(&busy_wait);
	return 0;
}

static void bioshd_geninit() 
{
  int i;

  for (i=0; i < 4<<6; i++) {
    if ((i &((1<<6)-1)) == 0) {
      hd[i].start_sect=0;
      hd[i].nr_sects=drive_info[i>>6].sectors*
	drive_info[i>>6].heads*
	  drive_info[i>>6].cylinders;
    } else {
      hd[i].start_sect=-1;
      hd[i].nr_sects=0;
    }
  }
  blksize_size[MAJOR_NR]=1024;
}
